#!/usr/bin/env python3
"""
Windows Event Log Viewer (System Management Tool)

Parameters:
    --list: List available event logs
    --view LOG: View events from specified log
    --filter TYPE: Filter by event type
    --since TIME: Show events since time
    --export FILE: Export events to file
    --clear LOG: Clear specified event log
    --format FORMAT: Output format (table/json)

Requires: win32evtlog, win32con, tabulate packages and admin privileges
Warning: Clearing event logs will permanently delete event history
"""

import os
import sys
import argparse
import logging
from typing import List, Dict, Any, Optional, Generator
from tabulate import tabulate
import win32evtlog
import win32evtlogutil
import win32security
import win32con
import wmi
import json
from datetime import datetime, timedelta
import time
import queue
import threading

class EventLogViewer:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.wmi = wmi.WMI()
        self.event_types = {
            win32evtlog.EVENTLOG_AUDIT_SUCCESS: 'AUDIT_SUCCESS',
            win32evtlog.EVENTLOG_AUDIT_FAILURE: 'AUDIT_FAILURE',
            win32evtlog.EVENTLOG_INFORMATION_TYPE: 'INFORMATION',
            win32evtlog.EVENTLOG_WARNING_TYPE: 'WARNING',
            win32evtlog.EVENTLOG_ERROR_TYPE: 'ERROR'
        }

    def get_events(self, log_type: str = 'System', 
                  event_types: Optional[List[str]] = None,
                  source: Optional[str] = None,
                  start_time: Optional[datetime] = None,
                  end_time: Optional[datetime] = None,
                  max_events: int = 100) -> List[Dict[str, Any]]:
        """Get events from specified log."""
        events = []
        handle = win32evtlog.OpenEventLog(None, log_type)
        flags = win32evtlog.EVENTLOG_BACKWARDS_READ | \
                win32evtlog.EVENTLOG_SEQUENTIAL_READ
        
        try:
            while len(events) < max_events:
                events_batch = win32evtlog.ReadEventLog(
                    handle, flags, 0
                )
                
                if not events_batch:
                    break
                    
                for event in events_batch:
                    if len(events) >= max_events:
                        break
                        
                    event_time = datetime.fromtimestamp(
                        int(event.TimeGenerated)
                    )
                    
                    # Apply time filters
                    if start_time and event_time < start_time:
                        continue
                    if end_time and event_time > end_time:
                        continue
                        
                    # Apply event type filter
                    if event_types:
                        event_type = self.event_types.get(
                            event.EventType, 'UNKNOWN'
                        )
                        if event_type not in event_types:
                            continue
                            
                    # Apply source filter
                    if source and event.SourceName != source:
                        continue
                        
                    events.append(self._format_event(event))
                    
        except Exception as e:
            self.logger.error(f"Error reading event log: {str(e)}")
            
        finally:
            win32evtlog.CloseEventLog(handle)
            
        return events

    def monitor_events(self, log_type: str = 'System',
                      event_types: Optional[List[str]] = None,
                      source: Optional[str] = None) -> Generator[Dict[str, Any], None, None]:
        """Monitor event log in real-time."""
        flags = win32evtlog.EVENTLOG_FORWARDS_READ | \
                win32evtlog.EVENTLOG_SEQUENTIAL_READ
        
        handle = win32evtlog.OpenEventLog(None, log_type)
        try:
            while True:
                events = win32evtlog.ReadEventLog(handle, flags, 0)
                
                if not events:
                    time.sleep(1)  # Wait before checking for new events
                    continue
                    
                for event in events:
                    # Apply event type filter
                    if event_types:
                        event_type = self.event_types.get(
                            event.EventType, 'UNKNOWN'
                        )
                        if event_type not in event_types:
                            continue
                            
                    # Apply source filter
                    if source and event.SourceName != source:
                        continue
                        
                    yield self._format_event(event)
                    
        except Exception as e:
            self.logger.error(f"Error monitoring event log: {str(e)}")
            
        finally:
            win32evtlog.CloseEventLog(handle)

    def search_events(self, query: str, log_type: str = 'System',
                     max_events: int = 100) -> List[Dict[str, Any]]:
        """Search events using a text query."""
        events = []
        handle = win32evtlog.OpenEventLog(None, log_type)
        flags = win32evtlog.EVENTLOG_BACKWARDS_READ | \
                win32evtlog.EVENTLOG_SEQUENTIAL_READ
        
        try:
            while len(events) < max_events:
                events_batch = win32evtlog.ReadEventLog(
                    handle, flags, 0
                )
                
                if not events_batch:
                    break
                    
                for event in events_batch:
                    if len(events) >= max_events:
                        break
                        
                    event_data = self._format_event(event)
                    
                    # Search in all text fields
                    searchable_text = ' '.join([
                        str(event_data['source']),
                        str(event_data['message']),
                        str(event_data['event_id']),
                        str(event_data['event_type'])
                    ]).lower()
                    
                    if query.lower() in searchable_text:
                        events.append(event_data)
                        
        except Exception as e:
            self.logger.error(f"Error searching event log: {str(e)}")
            
        finally:
            win32evtlog.CloseEventLog(handle)
            
        return events

    def get_log_statistics(self, log_type: str = 'System',
                          days: int = 7) -> Dict[str, Any]:
        """Get statistics about event log entries."""
        start_time = datetime.now() - timedelta(days=days)
        events = self.get_events(
            log_type=log_type,
            start_time=start_time,
            max_events=10000  # Large number to get good statistics
        )
        
        stats = {
            'total_events': len(events),
            'by_type': {},
            'by_source': {},
            'by_day': {},
            'most_common_events': {}
        }
        
        for event in events:
            # Count by type
            event_type = event['event_type']
            stats['by_type'][event_type] = \
                stats['by_type'].get(event_type, 0) + 1
                
            # Count by source
            source = event['source']
            stats['by_source'][source] = \
                stats['by_source'].get(source, 0) + 1
                
            # Count by day
            day = event['time'].strftime('%Y-%m-%d')
            stats['by_day'][day] = \
                stats['by_day'].get(day, 0) + 1
                
            # Count event IDs
            event_id = event['event_id']
            stats['most_common_events'][event_id] = \
                stats['most_common_events'].get(event_id, 0) + 1
                
        # Sort and limit most common events
        stats['most_common_events'] = dict(
            sorted(
                stats['most_common_events'].items(),
                key=lambda x: x[1],
                reverse=True
            )[:10]
        )
        
        return stats

    def _format_event(self, event) -> Dict[str, Any]:
        """Format a Windows event log entry."""
        try:
            return {
                'time': datetime.fromtimestamp(int(event.TimeGenerated)),
                'source': event.SourceName,
                'event_id': event.EventID & 0xFFFF,  # Mask out qualifiers
                'event_type': self.event_types.get(
                    event.EventType, 'UNKNOWN'
                ),
                'category': event.EventCategory,
                'message': event.StringInserts 
                          if event.StringInserts else [],
                'computer': event.ComputerName,
                'sid': self._get_sid_string(event.Sid) if event.Sid else None
            }
        except Exception as e:
            self.logger.debug(f"Error formatting event: {str(e)}")
            return {}

    def _get_sid_string(self, sid) -> str:
        """Convert a SID to string format."""
        try:
            return win32security.ConvertSidToStringSid(sid)
        except Exception:
            return str(sid)

    def display_results(self, data: Any, output_format: str = 'table'):
        """Display results in specified format."""
        if output_format == 'json':
            print(json.dumps(data, indent=4, default=str))
        else:
            self._display_table_format(data)

    def _display_table_format(self, data: Any):
        """Display data in table format."""
        if isinstance(data, list):
            if data and isinstance(data[0], dict):
                # For events, use a specific format
                headers = ['Time', 'Source', 'Event ID', 'Type', 'Message']
                table_data = []
                for event in data:
                    message = event.get('message', [])
                    if isinstance(message, list):
                        message = ' '.join(message)
                    table_data.append([
                        event['time'],
                        event['source'],
                        event['event_id'],
                        event['event_type'],
                        message[:100] + '...' if len(message) > 100 else message
                    ])
                print(tabulate(table_data, headers=headers, tablefmt='grid'))
        elif isinstance(data, dict):
            for key, value in data.items():
                if isinstance(value, (dict, list)):
                    print(f"\n{key}:")
                    if isinstance(value, dict):
                        for k, v in value.items():
                            print(f"  {k}: {v}")
                    else:  # list
                        for item in value:
                            print(f"  {item}")
                else:
                    print(f"{key}: {value}")

def main():
    parser = argparse.ArgumentParser(
        description='Event Log Viewer - Windows event log utility')
    parser.add_argument('--log-type', choices=['System', 'Application', 'Security'],
                       default='System',
                       help='Event log to view')
    parser.add_argument('--event-types', nargs='+',
                       choices=['INFORMATION', 'WARNING', 'ERROR', 
                               'AUDIT_SUCCESS', 'AUDIT_FAILURE'],
                       help='Event types to include')
    parser.add_argument('--source',
                       help='Filter by event source')
    parser.add_argument('--start-time',
                       help='Start time (YYYY-MM-DD HH:MM:SS)')
    parser.add_argument('--end-time',
                       help='End time (YYYY-MM-DD HH:MM:SS)')
    parser.add_argument('--max-events', type=int, default=100,
                       help='Maximum number of events to retrieve')
    parser.add_argument('--monitor', action='store_true',
                       help='Monitor events in real-time')
    parser.add_argument('--search',
                       help='Search for events containing text')
    parser.add_argument('--stats', action='store_true',
                       help='Show event log statistics')
    parser.add_argument('--days', type=int, default=7,
                       help='Number of days for statistics')
    parser.add_argument('--format', choices=['table', 'json'],
                       default='table',
                       help='Output format')
    
    args = parser.parse_args()
    
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    try:
        viewer = EventLogViewer()
        
        # Parse time arguments if provided
        start_time = None
        if args.start_time:
            start_time = datetime.strptime(
                args.start_time, '%Y-%m-%d %H:%M:%S'
            )
            
        end_time = None
        if args.end_time:
            end_time = datetime.strptime(
                args.end_time, '%Y-%m-%d %H:%M:%S'
            )
            
        if args.monitor:
            print(f"Monitoring {args.log_type} event log...")
            try:
                for event in viewer.monitor_events(
                    args.log_type, args.event_types, args.source
                ):
                    viewer.display_results([event], args.format)
            except KeyboardInterrupt:
                print("\nMonitoring stopped by user")
                
        elif args.search:
            events = viewer.search_events(
                args.search, args.log_type, args.max_events
            )
            viewer.display_results(events, args.format)
            
        elif args.stats:
            stats = viewer.get_log_statistics(
                args.log_type, args.days
            )
            viewer.display_results(stats, args.format)
            
        else:
            events = viewer.get_events(
                log_type=args.log_type,
                event_types=args.event_types,
                source=args.source,
                start_time=start_time,
                end_time=end_time,
                max_events=args.max_events
            )
            viewer.display_results(events, args.format)
            
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}")
        return 1
        
    return 0

if __name__ == "__main__":
    exit(main())
